
\begin{code}
module AI (think) where

import List
import Maybe

import Board
import Types
import Knowledge
import Dijkstra

think :: Knowledge -> (Bid, Command, Knowledge)
think k = (bid, c, k')
    where b = board k
          mp = my_pos k
          ps_dists = ps_here k
          carrying_packages = my_packages k
          droppable_packages = filter ((mp ==) . package_pos) carrying_packages
          sorted_packages = sortBy compare_package_tuple ps_dists
          pickable_packages = can_pick (carrying k) (max_weight k) sorted_packages
          (c, k')
            = if null droppable_packages
              then if null pickable_packages
                   then if backing_off k
                        then (Dunno, k)
                        else let ppls' = map fst $ ppls k
                                 ppls_lp' = map fst $ ppls_lp k
                             in if continuing k
                                then next_step' k b ppls_lp' mp
                                else if null carrying_packages
                                     then if null ppls'
                                          then if null ppls_lp'
                                               then let (c', k_with_dh) = best_step' k b ppls_lp' mp (map bot_pos $ filter ((> 0) . bot_count) (others k))
                                                    in (c', no_dh k_with_dh)
                                               else best_step' k b ppls_lp' mp ppls_lp'
                                          else best_step' k b ppls_lp' mp ppls'
                                     else (best_step' k b ppls_lp' mp $ map package_pos carrying_packages)
                   else (Pick (map package_id pickable_packages), no_dh k)
              else (Drop (map package_id droppable_packages), no_dh k)

          bid = 1

can_pick :: Weight -> Weight -> [(Package, Distance)] -> [Package]
can_pick _ _ [] = []
can_pick w m ((p,_):ps) = if w' <= m then p:can_pick w' m ps else can_pick w m ps
    where w' = w + package_weight p

-- We want to visit vs as they may have packages on

best_step' :: Knowledge -> Board -> [Position] -> Position -> [Position] -> (Command, Knowledge)
best_step' k b vs p ps = move k $ best_step b vs (dh k) p ps

next_step' :: Knowledge -> Board -> [Position] -> Position -> (Command, Knowledge)
next_step' k b vs p = move k (next_step b vs p (fromJust $ dh k))

move :: Knowledge -> Maybe (Direction, DijkstraHeading) -> (Command, Knowledge)
move k (Just (d, x)) = (Move d, k { dh = Just x, continuing = True })
move k Nothing = (Dunno, k)

compare_package_tuple :: (Package, Distance) -> (Package, Distance) -> Ordering
compare_package_tuple (p1, dist1) (p2, dist2) = (f (package_weight p2) * f dist1) `compare` (f (package_weight p1) * f dist2)
    where f :: (Integral a) => a -> Integer
          f = fromIntegral
\end{code}

