> module Adrenaline (on_adrenaline, adrenaline_reasoning)  where

> import Board
> import Types
> import Knowledge
> import Dijkstra (distanceToPointsSimple,distanceToPointsLimited)
> import Maybe

use distance_to_points to check the distance from other robots, as
well as the distances of the surrounding fields to the nearest
delivery points.

> on_adrenaline :: Knowledge -> Bool
> on_adrenaline s = danger b pos enemypos > threshold where
>     -- our knowledge
>     pos = my_pos s
>     enemypos = map bot_pos $ others s
>     b = board s
>     -- danger threshold
>     threshold = 0.4

> danger :: Board -> Position -> [Position] -> Float
> danger b pos enemypos = dngr
>     where
>     -- fold everything to a single danger value
>     dngr = sum $ map (((1.0::Float)/) . fromIntegral) enemydist
>     -- only the enemies that are actually reachable interest us.
>     enemydist = strip_nothing $ distanceToPointsLimited b radius pos enemypos
>     -- constants
>     radius = 10
>     -- helpers
>     strip_nothing = map (\a -> (fromJust a)) . (filter isJust)

the local configuration contains everything of importance for
adrenaline mode.

> data LocalConf
>   = LC { lc_alive    :: Bool,         -- are we still there?
>          lc_pos      :: Position,     -- our position
>          lc_dist     :: Int,          -- distance to dpos from root pos
>          lc_enemypos :: [Position],   -- the enemy positions
>          lc_board    :: Board,        -- the board
>          lc_dest     :: Position      -- where we want to go
>        }

adrenaline reasoning

TODO: execute the stored command if we reach one of our destinations.

> adrenaline_reasoning :: Knowledge -> (Bid, Command)
> adrenaline_reasoning k = (1,c)
>     where
>     -- build a local conf that represents our current knowledge
>     lc = init_lc k
>     -- build a list of any direction that is reachable
>     dirs = reachable_directions (board k) (lc_pos lc)
>     -- build a list of localconfs to consider
>     lcs = zipWith lc_move_dir (repeat lc) dirs
>     -- score this list
>     dirscores = zip dirs $ map alpha_beta_search lcs
>     -- find the highest-scored direction
>     dir = find_best_dir dirscores
>         where
>         find_best_dir :: [(Direction,Int)] -> Direction
>         find_best_dir (x:xs) = iter x xs
>             where
>             iter :: (Direction,Int) -> [(Direction,Int)] -> Direction
>             iter a []     = fst a
>             iter a (x:xs) = if snd x > snd a
>                             then iter x xs
>                             else iter a xs
>     -- build a command from it
>     c = Move dir

transfer information from knowledge to localconf

> init_lc :: Knowledge -> LocalConf
> init_lc k = LC { lc_alive    = True,
>                  lc_pos      = my_pos k,
>                  lc_dist     = dist,
>                  lc_enemypos = map bot_pos $ others k,
>                  lc_board    = board k,
>                  lc_dest     = dest
>                }
>             where
>             dist = distance_to_point (board k) (my_pos k) dest
>             dest = my_strip_maybe (dh k)
>             my_strip_maybe (Just v) = let (p,_,_) = v in p
>             my_strip_maybe Nothing = (0,0) -- FUCK!


> distance_to_point :: Board -> Position -> Position -> Int
> distance_to_point b p p' = if length distlist == 0
>                            then maxBound
>                            else head distlist  
>     where
>     distlist = catMaybes $ distanceToPointsSimple b p [p']

move the postman to a given position (assumes it is reachable)

> lc_move :: LocalConf -> Position -> LocalConf
> lc_move l p = l { lc_pos = p,
>                   lc_enemypos = update_enemy_pos (lc_board l) (lc_enemypos l) p (x',y')
>                 }
>     where
>     x' = fst p - fst oldpos
>     y' = snd p - snd oldpos
>     oldpos = lc_pos l
>     update_enemy_pos :: Board -> [Position] -> Position -> (Int,Int) -> [Position]
>     update_enemy_pos _ []     _ _ = []
>     update_enemy_pos b (e:es) p d = if e == p
>                                     then (ex+x',ey+y'):rest
>                                     else e:rest
>         where
>         rest = update_enemy_pos b es p d
>         (ex,ey) = e
>         (x',y') = d

move the postman in a given direction (assumes it is reachable)

TODO: make this recursive, fix.

> lc_move_dir :: LocalConf -> Direction -> LocalConf
> lc_move_dir l d = lc_move l $ next_pos (lc_pos l) d

calculate every possible local configuration that is reachable in a
single step

> our_moves :: LocalConf -> [LocalConf]
> our_moves l = iter l $ reachable_directions (lc_board l) (lc_pos l)
>     where
>     iter :: LocalConf -> [Direction] -> [LocalConf]
>     iter _ []     = []
>     iter l (p:ps) = lc_move_dir l p : iter l ps

helper code to do the permutation

> selections []     = []
> selections (x:xs) = (x,xs) : [ (y,x:ys) | (y,ys) <- selections xs ]
> permutations    :: [a] -> [[a]]
> permutations [] = [[]]
> permutations xs = [ y : zs | (y,ys) <- selections xs, zs <- permutations ys]

generate every possible enemy position. to do that, permute the list
of enemy locations, then spit out up to four localconfs for each.

> enemy_moves :: LocalConf -> [LocalConf]
> enemy_moves c = [] -- defunct
>    where
>    enemies :: [[Position]]
>    enemies = permutations $ lc_enemypos c

the alpha-beta search routine

> alpha_beta_search :: LocalConf -> Int
> alpha_beta_search c = evalmax c depth (minBound, maxBound)
>     where
>     depth = 10

evalmin evaluates a min node, under the restriction that it must
return a value in [min,max]. so if it, during evaluation, finds a
value smaller than min, it would evaluate to at most that value. it
may not do that, so instead it will return min then. additionally,
any time it sees a value that is smaller than max, it updates max to
that value.

> evalmin :: LocalConf -> Int -> (Int,Int) -> Int
> evalmin c 0 _ = score c
> evalmin c n b = iter (enemy_moves c) n b
>     where
>     iter :: [LocalConf] -> Int -> (Int,Int) -> Int
>     iter []     _ b = snd b
>     iter (c:cs) n b = if branchval < fst b
>                       then fst b
>                       else if branchval < snd b
>                            then iter cs n (fst b,branchval)
>                            else iter cs n b
>         where
>         branchval :: Int
>         branchval = evalmax c (n-1) b

evalmax works analogue to the above.

> evalmax :: LocalConf -> Int -> (Int,Int) -> Int
> evalmax c 0 _ = score c
> evalmax c n b = iter (our_moves c) n b
>     where
>     iter :: [LocalConf] -> Int -> (Int,Int) -> Int
>     iter []     _ b = fst b
>     iter (c:cs) n b = if branchval > snd b
>                       then snd b
>                       else if branchval > fst b
>                            then iter cs n (branchval,snd b)
>                            else iter cs n b
>         where
>         branchval :: Int
>         branchval = evalmin c (n-1) b

assign a certain score to a local configuration

> score :: LocalConf -> Int
> score c = if is_lethal (lc_board c) (lc_pos c)
>           then (- maxBound)
>           else progress_score - danger_score
>     where
>     -- knowledge from the local conf
>     board = lc_board c
>     pos   = lc_pos c
>     dest  = lc_dest c
>     enemypos  = lc_enemypos c
>     root_dist = lc_dist c
>     -- give a fixed number of points if we got nearer to the destination
>     destdist = distance_to_point board pos dest
>     progress_score = if destdist < root_dist then 50 else 0
>     -- give us a penalty for nearby bots
>     danger_score :: Int
>     danger_score = truncate $ (danger board pos enemypos) * fear
>     -- how much do we fear other bots?
>     fear = 35
