> module Backoff (backoff_reasoning) where

> import Knowledge (Knowledge, is_robot_at, reboots, board, my_pos,
>                   back_off, backing_off, rng)
> import Types (Direction(..), Command(..), Bid)
> import Board (reachable_directions, next_pos)
> import List (partition)
> import Random (randomR)


How long to stay in back-off mode?

> back_off_turns :: Int
> back_off_turns = 3

Under what condition do we enter back-off mode?

> should_enter :: Knowledge -> Bool
> should_enter k = reboots k 5 >= 3


> backoff_reasoning :: Knowledge -> Maybe (Bid, Command, Knowledge)
> backoff_reasoning k = backoff_behav $ decide_backoff k

> decide_backoff :: Knowledge -> Knowledge
> decide_backoff k = if not (backing_off k) && should_enter k
>                    then back_off back_off_turns k
>                    else k


> backoff_behav :: Knowledge -> Maybe (Bid, Command, Knowledge)
> backoff_behav k = if doit 
>                   then Just (1, Move nice_dir, k')
>                   else Nothing
>   where
>   doit = backing_off k && not (null dir_choices)
>
>   mp = my_pos k
>   (i,rng') = if doit then randomR (0,length dir_choices-1) (rng k)
>              else (0,(rng k))
>
>   k' = k { rng = rng' }
>
>   nice_dir :: Direction
>   nice_dir = dir_choices !! i
>
>   (free_dirs, bot_dirs) = partition has_bot
>                           (reachable_directions (board k) mp)
>
>   dir_choices :: [Direction]
>   dir_choices = free_dirs ++ bot_dirs
>
>   has_bot :: Direction -> Bool
>   has_bot d = is_robot_at k $ next_pos mp d

TODO: Don't move onto spaces adjacent to water.
