> module DistanceMap (DistanceMap, make_distance_map) where

> import Types
> import Board

> import Data.Array
> import Int

> type DistanceMap = Array Position Int16

the distance map is associated with a certain position. it maps a
position to the distance to associated position. for saving space, the
distance is a short.

so the algorithm starts at the given position. an initial array is
constructed, holding a special value to see if a position that is
encountered has already been seen. a region grower is then started. it
uses a variable to see how far from the seed we actually are. the
position list initially just contains the seed position, and in the
recursion it will always contain the indices of the fields that are
reached. imagine a circle that increases its radius by one in each
recursion step.

> make_distance_map :: Board -> Position -> DistanceMap
> make_distance_map b p = regiongrow b 0 [p] $
>                         listArray (size b) $ repeat maxval where
>
>     -- if the list is empty, every reachable field has a valid distance
>     regiongrow :: Board -> Int16 -> [Position] -> DistanceMap
>                   -> DistanceMap
>     regiongrow _ _ [] m = m
>     regiongrow b n p  m = regiongrow b (n+1) (newlist b m' p) m'
>         where m' = insertlist n p m
>
>     -- write the distance into all the unseen elements of the list
>     insertlist :: Int16 -> [Position] -> DistanceMap -> DistanceMap
>     insertlist _ []     m = m
>     insertlist n (p:ps) m = if unseen m p
>                             then insertlist n ps $ m//[(p,n)]
>                             else insertlist n ps m
>
>     -- build a list of new positions
>     newlist :: Board -> DistanceMap -> [Position] -> [Position]
>     newlist _ _ []     = []
>     newlist b m (p:ps) = filter (unseen m) $ reachable_neighbors b p
>                          ++ newlist b m ps
>
>     -- this is meant to approximate +infinity
>     maxval = maxBound::Int16
>
>     -- to check for new fields
>     unseen m p = m!p == maxval
